"""tetgen package version.

On the ``main`` branch, use 'dev0' to denote a development version.
For example:

version_info = 0, 27, 'dev0'

"""

# major, minor, patch, -extra
version_info = 0, 8, "dev0"

# Nice string for the version
__version__ = ".".join(map(str, version_info))
